package org.netkernel.lang.ncode;

import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.impl.HDSPredicateFactory;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.urii.impl.Version;

import java.util.*;

public class AvailablePalettesAccessor extends StandardAccessorImpl
{
	public void onSource(INKFRequestContext aContext) throws Exception
	{	AvailablePalettesRepresentation result = new AvailablePalettesRepresentation();
		INKFRequest req=aContext.createRequest("active:spaceAggregateHDS");
		req.addArgument("uri", "res:/etc/system/CDEPalette.xml");
		req.setRepresentationClass(IHDSNode.class);
		IHDSNode hds=(IHDSNode)aContext.issueRequest(req);
		IHDSNodeList spaces=hds.getNodes("/spaces/space").filter(HDSPredicateFactory.hasChild("palette"));
		
		//filter to only include latest version of each space
		Map<String,IHDSNode> spacesById=new HashMap<String,IHDSNode>();
		for(IHDSNode spaceNode : spaces)
		{	String id=spaceNode.getFirstValue("id").toString().trim();
			IHDSNode found=spacesById.get(id);
			if (found==null)
			{	spacesById.put(id, spaceNode);
			}
			else
			{	Version v1=new Version(spaceNode.getFirstValue("version").toString().trim());
				Version v2=new Version(found.getFirstValue("version").toString().trim());
				if (v1.isGreaterThan(v2))
				{	spacesById.put(id, spaceNode);
				}
			}
		}
		
		for(IHDSNode spaceNode : spacesById.values())
		{	IHDSNode paletteNode=spaceNode.getFirstNode("palette");
			for (IHDSNode itemNode: paletteNode.getChildren())
			{	
				String nodeName=itemNode.getName();
				if (nodeName.equals("group"))
				{	result.addGroup(itemNode);
				}
				else if (nodeName.equals("endpoint"))
				{	result.addItem(spaceNode,itemNode);
				}
			}
		}
		aContext.createResponseFrom(result);
	}
}
