package org.netkernel.lang.ncode;

import java.util.*;

import org.netkernel.ext.system.representation.IRepDeployedModules;
import org.netkernel.layer0.nkf.INKFLocale;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.representation.impl.HDSNodeListImpl;
import org.netkernel.layer0.util.MultiMap;
import org.netkernel.layer0.util.OrderedCheapMultiStringMap;

public class AvailablePalettesRepresentation
{
	List<IHDSNode> mGroups = new ArrayList<IHDSNode>();
	MultiMap mItemsBySpace = new MultiMap(100,16);
	
	public AvailablePalettesRepresentation()
	{
	}
	
	public void addGroup(IHDSNode aGroupNode)
	{	mGroups.add(aGroupNode);
	}

	public void addItem(IHDSNode aSpaceNode, IHDSNode aItemNode)
	{	mItemsBySpace.put(aSpaceNode,aItemNode);
	}
	
	public IHDSNodeList getSpaces()
	{	List<IHDSNode> nodes=new ArrayList<IHDSNode>(64);
		for (Iterator i=mItemsBySpace.keyIterator(); i.hasNext(); )
		{	IHDSNode spaceNode=(IHDSNode)i.next();
			nodes.add(spaceNode);
		}
		return new HDSNodeListImpl(nodes);
	}
	
	public void populatePalette(PaletteRepresentation aPalette, IHDSNodeList aChosenSpaces, IRepDeployedModules aModules, INKFRequestContext aContext) throws Exception
	{	
		for (IHDSNode groupNode: mGroups)
		{
			String id=(String)groupNode.getFirstValue("@id");
			String name=(String)groupNode.getFirstValue("@name");
			String desc=(String)groupNode.getFirstValue("@desc");
			String orderString=(String)groupNode.getFirstValue("@order");
			int order;
			if (orderString!=null)
			{	order=Integer.parseInt(orderString);
			}
			else
			{	order=Integer.MAX_VALUE-1000;
			}
			aPalette.addGroup(id, name, desc, order);
		}
		
		for (IHDSNode spaceNode : aChosenSpaces)
		{
			List<IHDSNode> items=mItemsBySpace.get(spaceNode);
			if (items!=null)
			{
				for (IHDSNode itemNode : items)
				{
					String id=(String)itemNode.getFirstValue("@id");
					String group=(String)itemNode.getFirstValue("@group");
					if (group==null)
					{	group="";
					}
					String orderString=(String)itemNode.getFirstValue("@order");
					int order=Integer.MAX_VALUE;
					if (orderString!=null)
					{	order=Integer.parseInt(orderString);
					}

					IHDSNodeList argNodes=itemNode.getNodes("argument");
					OrderedCheapMultiStringMap hardArgs=null;
					if (argNodes.size()>0)
					{	hardArgs=new OrderedCheapMultiStringMap(argNodes.size());
						for (IHDSNode argNode : argNodes)
						{	String argName=(String)argNode.getFirstValue("@name");
							String argValue=(String)argNode.getValue();
							hardArgs.put(argName, argValue);
						}
					}
					if (aModules.hashForSpaceElement(id)==null)
					{	aContext.logFormatted(INKFLocale.LEVEL_WARNING, "MSG_NCODE_PALETTE_NO_ITEM", id);
					}
					else
					{	LogicalEndpointPaletteItem item = new LogicalEndpointPaletteItem(id, hardArgs, order, aModules, aContext);
						IHDSNodeList injectedNodes=itemNode.getNodes("injected");
						if (injectedNodes.size()>0)
						{	HDSBuilder b=new HDSBuilder();
							b.pushNode("injected__A");
							for (IHDSNode injectedNode : injectedNodes)
							{	b.importNode(injectedNode);
							}
							item.setAdditionalOutputs(b.getRoot());
						}
						
						String name=(String)itemNode.getFirstValue("@name");
						if (name!=null)
						{	item.overrideName(name);
						}
						String idp=(String)itemNode.getFirstValue("@idp");
						if (idp!=null)
						{	item.overrideId(idp);
						}
						String cssClass=(String)itemNode.getFirstValue("@class");
						if (cssClass!=null)
						{	item.setClass(cssClass);
						}
						aPalette.addEndpoint(group, item, aContext);
					}
				}
			}
		}
		
		
	}
	
	
}
