package org.netkernel.lang.ncode;

import java.util.*;
import org.netkernel.container.IMessages;
import org.netkernel.layer0.bnf.BNFGrammarFactory;
import org.netkernel.layer0.meta.GrammarException;
import org.netkernel.layer0.meta.IEndpointArgumentMeta;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.IIdentifierGrammar;
import org.netkernel.layer0.meta.impl.EndpointArgumentMetaImpl;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.util.MultiMap;
import org.netkernel.layer0.util.OrderedCheapMultiStringMap;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.urii.IIdentifier;
import org.netkernel.xml.xda.DOMXDA;
import org.netkernel.xml.xda.IXDAReadOnly;
import org.netkernel.xml.xda.IXDAReadOnlyIterator;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

public class EndpointMetaWrapper
{
	private final IEndpointMeta mMeta;
	private final List<ArgDef> mArgDefs;
	//private final ArgDef mRenamableArgDef;
	private final IHDSNode mInitialArgs;
	private final OrderedCheapMultiStringMap mHardArgs;
	private final DOMXDA mGrammarXDA;
	private final boolean mIsActive;

	public final static IHDSNode sNoArgs;
	public final static OrderedCheapMultiStringMap sNoHardArgs;
	static
	{
		HDSBuilder b = new HDSBuilder();
		sNoArgs = b.getRoot();
		sNoHardArgs = new OrderedCheapMultiStringMap(0);
	}
	
	public EndpointMetaWrapper(IEndpointMeta aMeta, OrderedCheapMultiStringMap aHardwiredArgs, INKFRequestContext aContext) throws Exception
	{
		mHardArgs=aHardwiredArgs!=null?aHardwiredArgs:sNoHardArgs;
		mMeta = aMeta;
		IIdentifierGrammar grammar = mMeta.getIdentifierGrammar();
		Document d=aContext.transrept(grammar.toXMLString(), Document.class);
		mGrammarXDA=new DOMXDA(d,false);
		mIsActive = mGrammarXDA.isTrue("/group[group[1]/@name='scheme' and *[2]/@name='activeType']");
		if (mIsActive)
		{	mArgDefs=getActiveArgDefs();
		}
		else
		{	mArgDefs=getRegularArgDefs();
		}
		/*
		ArgDef renamableArgDef=null;
		for (ArgDef def : mArgDefs)
		{	if (def.mRenamable)
			{	renamableArgDef=def;
				break;
			}
		}
		mRenamableArgDef=renamableArgDef;
		*/
		mInitialArgs = getArguments(sNoArgs);		
	}
	
	private List<ArgDef> getRegularArgDefs() throws Exception
	{
		List<ArgDef> argDefs = new ArrayList<ArgDef>();
		for (IXDAReadOnlyIterator i = mGrammarXDA.readOnlyIterator("//*[@name and (name()='group' or name()='optional' or name()='interleave' or name()='choice')]"); i.hasNext();)
		{	i.next();
			String argName=i.getText("@name", false);
			/*
			boolean canEval=i.isTrue("count(*)=1 and name(*[1])='regex' and not(text()) and *[1]/@type");
			//boolean canEval1=i.isTrue("count(*)=1");
			//boolean canEval2=i.isTrue("name(*[1])='regex' ");
			//boolean canEval3=i.isTrue("not(text())");
			String argType;
			if (canEval)
			{	argType=i.eval("*[1]/@type");
			}
			else
			{	argType="freeform";
			}
			*/
			/*
			if (argType.equals(BNFGrammarFactory.BUILTIN_ACTIVE_ESCAPED_URI_SIMPLE.getName()) || argType.equals(BNFGrammarFactory.BUILTIN_ACTIVE_ESCAPED_URI))
			{	
			}
			*/
			IEndpointArgumentMeta argMeta=getArgumentMeta(argName);
			ArgDef def=new ArgDef(argName, null, 0, 0, false, argName, argMeta);
			argDefs.add(def);
		}
		return argDefs;
	}
	
	private List<ArgDef> getActiveArgDefs() throws Exception
	{
		List<ArgDef> argDefs = new ArrayList<ArgDef>();
		//String activeArgGrammar=BNFGrammarFactory.BUILTIN_ACTIVE_ESCAPED_URI_SIMPLE.getName();
		for (IXDAReadOnlyIterator i = mGrammarXDA.readOnlyIterator("/group/interleave/group[not(last() and count(group)=2)]"); i.hasNext();)
		{	i.next();
			String name=i.getText("group/@name", true);
			int min,max;
			try
			{	String minString=i.getText("@min", true);
				min=Integer.parseInt(minString);
			} catch (Exception e)
			{	min=1;
			}
			try
			{	String maxString=i.getText("@max", true);
				max=maxString.equals("*")?Integer.MAX_VALUE:Integer.parseInt(maxString);
			}
			catch (Exception e)
			{	max=1;
			}
			IEndpointArgumentMeta argMeta=getArgumentMeta(name);
			ArgDef def=new ArgDef(name, null, min, max, false, name, argMeta);
			argDefs.add(def);
		}
		
		boolean hasVarargs=mGrammarXDA.isTrue("/group/interleave/group[last() and count(group)=2]");
		if (hasVarargs)
		{	IEndpointArgumentMeta argMeta=getArgumentMeta("varargs");
			String desc=null;
			String name=null;
			if (argMeta!=null)
			{	desc=argMeta.getDescription();
				name=argMeta.getDefaultValue();
			}
			if (desc==null) desc="additional arbitrary argument";
			if (name==null) name="vararg";
			ArgDef def=new ArgDef("vararg",desc, 0, Integer.MAX_VALUE, true, name, null);
			argDefs.add(def);
		}
		return argDefs;
	}
	
	
	
	public IEndpointMeta getMeta()
	{	return mMeta;
	}
	
	public IHDSNode getInitialArgs()
	{	return mInitialArgs;
	}
	
	public OrderedCheapMultiStringMap getHardWiredArgs()
	{	return mHardArgs;
	}
	
	public IHDSNode validateInputs(IHDSNode aInputs, IMessages aMessages)
	{
		IHDSNodeList inputs=aInputs.getNodes("i");
		IIdentifierGrammar grammar=mMeta.getIdentifierGrammar();
		OrderedCheapMultiStringMap parts=new OrderedCheapMultiStringMap(inputs.size());
		
		for (int i=0; i<mHardArgs.size(); i++)
		{	parts.put(mHardArgs.getKey(i), mHardArgs.getValue(i));
		}
		

		
		boolean valid=false;
		try
		{	
			
			for (IHDSNode input : inputs)
			{	
				if ((Boolean)input.getFirstValue("c"))
				{	String name=(String)input.getFirstValue("n");
					String proto=(String)input.getFirstValue("p");
					ArgDef def=getArgDefFor(proto);
					if (def.mRenamable)
					{	parts.put("argName", name);
						parts.put("argValue", "arg:"+name);
					}
					else
					{	parts.put(name, "arg:"+name);
					}
				}
			}			
			
			String identifier=grammar.construct(parts, aMessages);
			valid=true;
		}
		catch (Exception e)
		{
		}
		
		HDSBuilder b=new HDSBuilder();
		b.addNode("valid", valid);
		
		IHDSNode args=getArguments(aInputs);
		b.importChildren(args);

		return b.getRoot();
	}
	
	
	private IHDSNode getArguments(IHDSNode aExistingArgs)
	{	HDSBuilder b=new HDSBuilder();
		
		if (mIsActive)
		{	getArgumentsActive(b,aExistingArgs);
		}
		else
		{	getArgumentsRegular(b,aExistingArgs);
		}
		
		//remove any hard coded arguments from response
		if (mHardArgs.size()>0)
		{	Set<String> hardArgNames=new HashSet<String>();
			for (int i=0; i<mHardArgs.size(); i++)
			{	hardArgNames.add(mHardArgs.getKey(i));
			}
			for (IHDSNode inputNode: b.getRoot().getNodes("/input__A/input"))
			{
				String inputName=(String)inputNode.getFirstValue("p");
				if (hardArgNames.contains(inputName))
				{	b.setCursor(inputNode);
					b.deleteNode();
				}
			}
		}
		
		//optimise response to maximally reuse inputs
		IHDSNodeList inputsIn=aExistingArgs.getNodes("i");
		Set<Integer> usedInputs = new HashSet<Integer>();
		for (IHDSNode inputNode: b.getRoot().getNodes("/input__A/input"))
		{	Integer io=(Integer)inputNode.getFirstValue("i");
			if (io!=null)
			{	usedInputs.add(io);
			}
		}
		for (IHDSNode inputNode: b.getRoot().getNodes("/input__A/input"))
		{	if (inputNode.getFirstValue("i")==null)
			{	String p=(String)inputNode.getFirstValue("p");
				int i=0;
				for (IHDSNode inputIn : inputsIn)
				{	String pIn=(String)inputIn.getFirstValue("p");
					Integer io=Integer.valueOf(i);
					if (pIn.equals(p) && !usedInputs.contains(io))
					{	usedInputs.add(io);
						b.setCursor(inputNode);
						b.addNode("i", io);
						break;
					}
					i++;
				}
			}
		}
		return b.getRoot();
	}
	
	private void getArgumentsRegular(HDSBuilder b, IHDSNode aExistingArgs)
	{
		MultiMap connected = new MultiMap(8, 4);
		//MultiMap unconnected = new MultiMap(8, 4);
		IHDSNodeList inputs=aExistingArgs.getNodes("i");
		int i=0;
		for (IHDSNode input : inputs)
		{	String proto=(String)input.getFirstValue("p");
			String name=(String)input.getFirstValue("n");
			if ((Boolean)input.getFirstValue("c"))
			{	connected.put(proto, new ExistingArg(name,i));
			}
			else
			{	
				//unconnected.put(proto, new ExistingArg(name,i));
			}
			i++;
		}
		
		b.pushNode("input__A");
		Element root=mGrammarXDA.getDocument().getDocumentElement();
		
		CountMap counts=new CountMap();
		recurseArgumentsRegular(root,b,connected,counts,false,false);
		
	}
	
	private int recurseArgumentsRegular(Element aGroupElement, HDSBuilder aBuilder, MultiMap aConnected, CountMap aCounts, boolean aForceAppear, boolean aForceMandatory)
	{
		int result=0;
		String name=aGroupElement.getAttribute("name");
		int min=aForceMandatory?getMin(aGroupElement):recurseGetMin(aGroupElement,null,aConnected);
		int max=recurseGetMax(aGroupElement,null,aConnected);
		if (min==0) aForceMandatory=false;
		if (name.length()>0)
		{
			int argIndex=aCounts.get(name);
			List<ExistingArg> existingArgs=aConnected.get(name);
			int required=Math.min(existingArgs.size()+1, max);
			if (argIndex<required || aForceAppear)
			{	
				ArgDef def=getArgDefFor(name);
				boolean optional=argIndex>=min && !aForceMandatory;
				if (argIndex<existingArgs.size())
				{	ExistingArg existing=existingArgs.get(argIndex);
					if (aBuilder!=null) writeInput(aBuilder, existing.mName, optional, existing.mOriginalIndex, def);
					result=2;
				}
				else
				{	if (aBuilder!=null) writeInput(aBuilder, name, optional, -1, def);
					result=1;
				}
				aCounts.increment(name);
			}
		}
		else
		{	String elementName=XMLUtils.pleaseDOMGiveMeTheNameOf(aGroupElement);
			
			if (elementName.equals("choice"))
			{	Element choice=getChosenChoice(aGroupElement,aConnected);
				int r=0;
				if (choice!=null)
				{	r=recurseArgumentsRegular(choice,aBuilder,aConnected,aCounts,false,false);
				}
				else
				{	for (Element e=XMLUtils.getFirstChildElement(aGroupElement); e!=null; e=XMLUtils.getNextSiblingElement(e))
					{	int r2=recurseArgumentsRegular(e,aBuilder,aConnected,aCounts,false,false);
						if (r2>r) r=r2;
					}
				}
				result=r;
			}
			else if (!elementName.equals("regex"))
			{	
				//determine number of iterations
				CountMap dryRunCounts=new CountMap(aCounts);
				int r;
				List<Integer> runResults=new ArrayList<Integer>();
				do
				{	r=0;
					for (Element e=XMLUtils.getFirstChildElement(aGroupElement); e!=null; e=XMLUtils.getNextSiblingElement(e))
					{	String childName=XMLUtils.pleaseDOMGiveMeTheNameOf(aGroupElement);
						if (!childName.equals("regex"))
						{	int r2=recurseArgumentsRegular(e,null,aConnected,dryRunCounts,false,false);
							if (r2>r) r=r2;
						}
					}
					runResults.add(r);
					if (r>result)
					{	result=r;
					}
				} while (r>0);

				for (int i=0; i<runResults.size(); i++)
				{	int rr=runResults.get(i);
					if (rr>0)
					{	for (Element e=XMLUtils.getFirstChildElement(aGroupElement); e!=null; e=XMLUtils.getNextSiblingElement(e))
						{	String childName=XMLUtils.pleaseDOMGiveMeTheNameOf(aGroupElement);
							if (!childName.equals("regex"))
							{	boolean forceMandatory=rr>1;
								recurseArgumentsRegular(e,aBuilder,aConnected,aCounts,true,forceMandatory);
							}
						}
					}
				}
			}				
		}
		return result;		
	}
	
	private int recurseGetMin(Element aGroupElement, Element aChild, MultiMap aConnected)
	{
		String minString=aGroupElement.getAttribute("min");
		int min;
		if (minString.length()>0)
		{	min=Integer.parseInt(minString);
		}
		else
		{	String elementName=XMLUtils.pleaseDOMGiveMeTheNameOf(aGroupElement);
			if (elementName.equals("optional"))
			{	min=0;
			}
			else if (elementName.equals("choice"))
			{	Element choice=getChosenChoice(aGroupElement,aConnected);
				min=(choice==null || choice==aChild)?1:0;
			}
			else
			{	min=1;
			}
		}
		Node parent=aGroupElement.getParentNode();
		if (parent instanceof Element)
		{
			int parentMin=recurseGetMin((Element)parent,aGroupElement,aConnected);
			min=min*parentMin;
		}
		return min;
	}
	
	private int getMin(Element aGroupElement)
	{
		String minString=aGroupElement.getAttribute("min");
		int min;
		if (minString.length()>0)
		{	min=Integer.parseInt(minString);
		}
		else
		{	String elementName=XMLUtils.pleaseDOMGiveMeTheNameOf(aGroupElement);
			if (elementName.equals("optional"))
			{	min=0;
			}
			else
			{	min=1;
			}
		}
		return min;
	}

	private int recurseGetMax(Element aGroupElement, Element aChild, MultiMap aConnected)
	{
		String maxString=aGroupElement.getAttribute("max");
		int max;
		if (maxString.length()>0)
		{	max=maxString.equals("*")?Integer.MAX_VALUE:Integer.parseInt(maxString);
		}
		else
		{	String elementName=XMLUtils.pleaseDOMGiveMeTheNameOf(aGroupElement);
			if (elementName.equals("optional"))
			{	max=1;
			}
			else if (elementName.equals("choice"))
			{	Element choice=getChosenChoice(aGroupElement,aConnected);
				max=(choice==null || choice==aChild)?1:0;
			}
			else
			{	max=1;
			}
		}
		Node parent=aGroupElement.getParentNode();
		if (parent instanceof Element)
		{
			int parentMax=recurseGetMax((Element)parent,aGroupElement,aConnected);
			max=max*parentMax;
		}
		
		return max;
	}
	
	private Element getChosenChoice(Element aChoiceElement, MultiMap aConnected)
	{	Element result=null;
		for (Element e=XMLUtils.getFirstChildElement(aChoiceElement); e!=null; e=XMLUtils.getNextSiblingElement(e))
		{
			for (Element e2=e; e2!=null; e2=XMLUtils.inOrderTraversalNext(e2, e))
			{	String name=e2.getAttribute("name");
				if (name.length()>0)
				{	if (aConnected.get(name).size()>0)
					{	result=e;
						break;
					}
				}
			}
			if (result!=null) break;
		}
		return result;
	}
	
	
	
	private void getArgumentsActive(HDSBuilder b, IHDSNode aExistingArgs)
	{
		MultiMap connected = new MultiMap(8, 4);
		IHDSNodeList inputs=aExistingArgs.getNodes("i");
		int i=0;
		for (IHDSNode input : inputs)
		{	String proto=(String)input.getFirstValue("p");
			String name=(String)input.getFirstValue("n");
			if ((Boolean)input.getFirstValue("c"))
			{	connected.put(proto, new ExistingArg(name,i));
			}
			i++;
		}
		
		b.pushNode("input__A");
		for (ArgDef argDef : mArgDefs)
		{	String name=argDef.mInitialName;
			List existing=connected.get(argDef.mName);
			int c=0;
			for (Object e : existing)
			{	ExistingArg existingArg=(ExistingArg)e;
				String named=existingArg.mName;
				boolean optional=c>=argDef.mMin;
				writeInput(b, named, optional, existingArg.mOriginalIndex, argDef);
				c++;
			}
			int required=Math.min(c+1, argDef.mMax);
			while(c<required)
			{	boolean optional=c>=argDef.mMin;
				writeInput(b, name, optional, -1, argDef);
				c++;
			}						
		}
		
	}
	
	private static void writeInput(HDSBuilder b, String aName, boolean aOptional, int aOriginalIndex, ArgDef aArgDef)
	{	b.pushNode("input");
		b.addNode("p", aArgDef.mName);
		b.addNode("n", aName);
		b.addNode("d", aArgDef.mDesc);
		b.addNode("t", aArgDef.mType.name());
		b.addNode("s", aArgDef.mTypeString);
		b.addNode("r", aArgDef.mRenamable);
		b.addNode("o", aOptional);
		if (aOriginalIndex>=0)
		{	b.addNode("i", aOriginalIndex);
		}
		b.popNode();
	}	
	
	private ArgDef getArgDefFor(String aName)
	{	ArgDef result=null;
		for (ArgDef argDef : mArgDefs)
		{	if (argDef.mName.equals(aName))
			{	result=argDef;
				break;
			}
		}
		return result;
	}	
	
	private IEndpointArgumentMeta getArgumentMeta(String aName)
	{	IEndpointArgumentMeta result=null;
		IEndpointArgumentMeta[] args=mMeta.getArguments();
		for (int i=0; i<args.length; i++)
		{	if (args[i].getName().equals(aName))
			{	result=args[i];
				break;
			}
		}
		return result;
	}	
		
	private static class ExistingArg
	{	public final String mName;
		public final int mOriginalIndex;
		public ExistingArg(String aName, int aOriginalIndex)
		{	mName=aName;
			mOriginalIndex=aOriginalIndex;
		}
	}
		
	private static class ExistingArg2
	{	public final String mName;
		public final int mOriginalIndex;
		public final boolean mConnected;
		public ExistingArg2(String aName, int aOriginalIndex, boolean aConnected)
		{	mName=aName;
			mOriginalIndex=aOriginalIndex;
			mConnected=aConnected;
		}
	}
		
	
	
	private static class ArgDef
	{	public final String mName;
		public final String mInitialName;
		public final String mDesc;
		public final int mMin;
		public final int mMax;
		public final boolean mRenamable;
		
		public final IEndpointArgumentMeta.ArgumentType mType;
		public final String mTypeString;
		
		public ArgDef(String aName, String aDesc, int aMin, int aMax, boolean aRenamable, String aInitialName, IEndpointArgumentMeta aMeta)
		{	mName=aName;
			if (aDesc!=null)
			{	mDesc=aDesc;
			}
			else if (aMeta!=null)
			{	mDesc=aMeta.getDescription();
			}
			else
			{	mDesc=null;
			}
			mMin=aMin;
			mMax=aMax;
			mRenamable=aRenamable;
			mInitialName=aInitialName;
			
			if (aMeta==null)
			{	mType=IEndpointArgumentMeta.ArgumentType.REPRESENTATION;
				mTypeString="no argument metadata defined";
			}
			else
			{	mType=aMeta.getArgumentType();
				switch(mType)
				{	case REPRESENTATION:
						mTypeString=aMeta.getRequiredRepresentations()[0].getName();
						break;
					case IDENTIFIER:
						mTypeString="";
						break;
					default:
					case STRING:
						mTypeString=aMeta.getSubstringType();
						break;
				}
			}			
		}
	}
}
