package org.netkernel.lang.ncode;

import java.io.ByteArrayInputStream;

import org.netkernel.container.ILogger;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.impl.EndpointMetaBuilder;
import org.netkernel.layer0.meta.impl.MetadataUtils;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.builtin.PrivateFilterEndpoint;
import org.w3c.dom.*;

public class Flow
{
	private final IEndpointMeta mMeta;
	private final String mId;
	private IHDSNode mConfig;
	
	public Flow(IHDSNode aConfig, INKFRequestContext aContext) throws Exception
	{	String grammar=(String)aConfig.getFirstValue("grammar");
		Document grammarDoc=XMLUtils.parse(new ByteArrayInputStream(grammar.getBytes("UTF-8")));
		String name=(String)aConfig.getFirstValue("name");
		String desc=(String)aConfig.getFirstValue("desc");
		EndpointMetaBuilder builder = new EndpointMetaBuilder();
		ILogger logger=aContext.getKernelContext().getKernel().getLogger();
		MetadataUtils.parseGrammar(grammarDoc.getDocumentElement(), builder, logger);
		builder.setName(name);
		builder.setDescription(desc);
		Integer verbs=(Integer)aConfig.getFirstValue("verbs");
		builder.setVerbs(verbs);
		if (!(Boolean)aConfig.getFirstValue("isPublic"))
		{	builder.addField(PrivateFilterEndpoint.KEY_PRIVATE, Boolean.TRUE);
		}
		mMeta=builder.getMeta(logger);
		mId=(String)aConfig.getFirstValue("id");
		mConfig=aConfig;
	}
	
	public String getId()
	{	return mId;
	}
	
	public IEndpointMeta getMeta()
	{	return mMeta;
	}
	
	public IHDSNode getConfig()
	{	return mConfig;
	}
	
}
