package org.netkernel.lang.ncode;

import org.netkernel.ext.system.representation.IRepDeployedModules;
import org.netkernel.lang.ncode.PaletteRepresentation.IPaletteItem;
import org.netkernel.lang.ncode.editor.IValueValidator;
import org.netkernel.layer0.meta.IEndpointArgumentMeta;
import org.netkernel.layer0.meta.IEndpointArgumentMeta.ArgumentType;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.impl.MetadataUtils;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.representation.IBinaryStreamRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.util.OrderedCheapMultiStringMap;
import org.netkernel.urii.IIdentifier;

public class LogicalEndpointPaletteItem implements IPaletteItem, Comparable<IPaletteItem>
{
	private final String mEndpointId;
	private final int mOrder;
	private final EndpointMetaWrapper mEndpointMeta;
	private String mIcon;
	private IHDSNode mAdditionalOutputs;
	private boolean mDynamic;
	private String mName;
	private String mPaletteId;
	private String mClass;
	
	public LogicalEndpointPaletteItem(String aId, OrderedCheapMultiStringMap aHardwiredArgs, int aOrder, IRepDeployedModules aModules, INKFRequestContext aContext) throws Exception
	{	mEndpointId = aId;
		mPaletteId=aId;
		mOrder = aOrder;
		String hash = aModules.hashForSpaceElement(mEndpointId);
		IEndpointMeta meta = (IEndpointMeta) aModules.metadataForHash(hash);
		mEndpointMeta=new EndpointMetaWrapper(meta, aHardwiredArgs, aContext);
		
		IIdentifier icon=mEndpointMeta.getMeta().getIcon();
		if (icon!=null)
		{	int i=hash.lastIndexOf('/');
			String spaceHash=hash.substring(0,i+1);
			mIcon="/tools/cr/"+spaceHash+icon.toString();
		}
		else
		{	mIcon=null;
		}
	}
	
	public void setAdditionalOutputs(IHDSNode aAdditionalOutputs)
	{	mAdditionalOutputs=aAdditionalOutputs;
	}
	
	public void setDynamic()
	{	mDynamic=true;
	}
	public void setClass(String aClass)
	{	mClass=aClass;
	}
	
	public void setIcon(String aIcon)
	{	mIcon=aIcon;
	}
	
	public void overrideName(String aName)
	{	mName=aName;
	}
	
	public void overrideId(String aId)
	{	mPaletteId=aId;
	}
	

	@Override
	public int compareTo(IPaletteItem o)
	{	return this.getOrder() - o.getOrder();
	}

	@Override
	public int getOrder()
	{	return mOrder;
	}

	@Override
	public String getEndpointId()
	{	return mEndpointId;
	}

	public String getPaletteId()
	{	return mPaletteId;
	}

	@Override
	public void serializeTo(HDSBuilder b)
	{	b.pushNode("endpoint");
		b.addNode("id", mPaletteId);
		b.addNode("eid", mEndpointId);
		b.addNode("name", mName!=null?mName:mEndpointMeta.getMeta().getName());
		b.addNode("desc", mEndpointMeta.getMeta().getDescription());
		b.addNode("valid", false);
		if (mDynamic) b.addNode("isDynamic",true);
		if (mClass!=null) b.addNode("cssClass",mClass);
		if (mIcon!=null)
		{	b.addNode("icon",mIcon);
		}
		b.importChildren(mEndpointMeta.getInitialArgs());
		b.pushNode("output");
		b.pushNode("response");
		b.addNode("t", IEndpointArgumentMeta.ArgumentType.REPRESENTATION.name());
		Class[] rca=mEndpointMeta.getMeta().getResponseClasses();
		if (rca!=null && rca.length>0)
		{	b.addNode("s",rca[0].getName());
		}
		else
		{	b.addNode("s","java.lang.Object");
		}
		b.popNode();
		
		if (mAdditionalOutputs!=null)
		{	b.importChildren(mAdditionalOutputs);
		}
		
		b.popNode(); //pop output
		b.popNode(); //pop endpoint
	}

	public IHDSNode validateInputs(IHDSNode aInputs, INKFRequestContext aContext) throws Exception
	{	return mEndpointMeta.validateInputs(aInputs, aContext.getKernelContext().getKernel().getLogger());
	}
	
	public IBinaryStreamRepresentation getIcon(INKFRequestContext aContext) throws Exception
	{	return null;
	}

	public IHDSNode validateValue(String aValue)
	{	return IValueValidator.VALID;
	}

	public OrderedCheapMultiStringMap getHardWiredInputs()
	{	return mEndpointMeta.getHardWiredArgs();
	}	
	
	/*
	public boolean isIdentifierInput(String aInputProto)
	{	IEndpointArgumentMeta meta=MetadataUtils.getEndpointIdentifierPart(mEndpointMeta.getMeta(),aInputProto);
		return meta==null || meta.isIdentifier();
	}
	*/

	public ArgumentType getInputType(String aInputName)
	{	IEndpointArgumentMeta meta=MetadataUtils.getEndpointIdentifierPart(mEndpointMeta.getMeta(),aInputName);
		if (meta==null)
		{	return ArgumentType.REPRESENTATION;
		}
		else
		{	return meta.getArgumentType();
		}
	}

	public String getDocumentation()
	{	return null;
	}	
	
}