package org.netkernel.lang.ncode;

import org.netkernel.container.IKernel;
import org.netkernel.ext.system.representation.IRepDeployedModules;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.impl.NKFTransportContextImpl;
import org.netkernel.layer0.representation.IBinaryStreamRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.representation.impl.HDSFactory;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.module.standard.endpoint.ConfiguredEndpointImpl;
import org.netkernel.module.standard.endpoint.StandardEndpointImpl;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.IResolution;
import org.netkernel.urii.ISpace;
import java.util.*;

public class NCODEPrototype extends ConfiguredEndpointImpl
{
	private INKFRequestContext mContext;
	private String mDynamicImportConfig;
	
	public NCODEPrototype()
	{	declareThreadSafe();
	}
	
	public void postCommission(INKFRequestContext aContext) throws Exception
	{	super.postCommission(aContext);
		mContext = new NKFTransportContextImpl(getKernel(),getSpace(),this,getParameters(),aContext.getKernelContext().getRequestScope());
	}
	
	public String toString()
	{	String name= "nCoDE Instance";
		try
		{	name=(String)getParameter("name");
		}
		catch (Exception e)	{}
		return name;
	}
	
	public ISpace getSpace()
	{	return super.getSpace();
	}
	
	protected IConfig createConfig(INKFRequestContext aContext) throws Exception
	{	IHDSNode config;
		try
		{	IReadableBinaryStreamRepresentation bs=aContext.source("param:operator",IReadableBinaryStreamRepresentation.class);
			config=HDSFactory.parseXML(bs.getInputStream());			
		}
		catch (Exception e)
		{	config=createDefaultConfig(aContext);
		}
		
		
		IIdentifier dynamicImportConfig=(IIdentifier)getParameter("dynamicImportConfig");
		mDynamicImportConfig=dynamicImportConfig!=null?dynamicImportConfig.toString():null;
		return new NCODEConfig(config,this,aContext,mDynamicImportConfig);
	}
	
	private IHDSNode createDefaultConfig(INKFRequestContext aContext) throws Exception
	{
		HDSBuilder b=new HDSBuilder();
		b.pushNode("cde");
		b.pushNode("flow__A");
		b.popNode();
		return b.getRoot();
	}
	

	protected void destroyConfig(IConfig aConfig) throws Exception
	{
	}
	
	public IKernel getKernel()
	{	return super.getKernel();
	}
	
	public NCODEConfig getDataFlowConfig(INKFRequestContext aContext) throws Exception
	{	return (NCODEConfig)getConfig(aContext);
	}
	
	public void save(IBinaryStreamRepresentation aState) throws Exception
	{	mContext.sink("param:operator", aState);
	}

	public Object load() throws Exception
	{	if (!mContext.exists("param:operator"))
		{	IHDSNode config=createDefaultConfig(mContext);
			return config;
			//return mContext.transrept(config, IReadableBinaryStreamRepresentation.class);
		}
		else
		{	return mContext.source("param:operator");
		}
	}
	
	/** determine verb support for operator resource
	 * @return verbs mask or -1 if operator not resolvable
	 */
	public int getOperatorVerbSupport() throws Exception
	{	int result=-1;
		INKFRequest req=mContext.createRequest("param:operator");
		IResolution resolution=mContext.getKernelContext().resolveRequest(req);
		if (resolution!=null)
		{	IIdentifier resolvedEndpoint=resolution.getEndpointIdentifier();
			if (resolvedEndpoint!=null)
			{	IMetaRepresentation m=mContext.meta(resolvedEndpoint.toString());
				if (m instanceof IEndpointMeta)
				{	IEndpointMeta epMeta=(IEndpointMeta)m;
				result=epMeta.getSupportedVerbs();
				}
			}
		}
		return result;
	}
	
	public Set<ISpace> findSpacesInScope(INKFRequestContext aContext, IRepDeployedModules aModules) throws Exception
	{
		Map<ISpace,ISpace> result=new IdentityHashMap<ISpace,ISpace>();
		IRequestScopeLevel scope=mContext.getKernelContext().getRequestScope();
		while (scope!=null)
		{	ISpace space=scope.getSpace();
			result.put(space,space);
			String spaceHash=aModules.hashForComponent(space);
			if (spaceHash!=null)
			{	ISpaceMeta spaceMeta=(ISpaceMeta)aModules.metadataForHash(spaceHash);
				ISpaceElements ise=spaceMeta.getElements();
				for (int i=0; i<ise.size(); i++)
				{	if (mDynamicImportConfig!=null)
					{	IEndpoint ep=ise.getPhysicalEndpoint(i);
						if (ep.getClass().getName().equals("org.netkernel.layer1.endpoint.DynamicImportEndpoint"))
						{	StandardEndpointImpl sep=(StandardEndpointImpl)ep;
							String diConfigId=sep.getParameter("config").toString();
							if (diConfigId.equals(mDynamicImportConfig))
							{	//skip any spaces found though the linked dynamic import
								continue;
							}
						}
					}
				
					ISpace delegatedSpace=ise.getDelegatedSpace(i);
					if (delegatedSpace!=null)
					{	result.put(delegatedSpace,delegatedSpace);
					}
				}
			}
			scope=scope.getParent();
		}
		return result.keySet();
	}
}