package org.netkernel.lang.ncode;

import org.json.JSONObject;
import org.netkernel.ext.system.representation.IRepDeployedModules;
import org.netkernel.lang.ncode.editor.ValueValidatorFactory;
import org.netkernel.layer0.meta.IEndpointArgumentMeta;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.IHDSPredicate;
import org.netkernel.layer0.representation.WrappedThrowable;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.util.CheapMap;
import org.netkernel.layer0.util.OrderedCheapMultiStringMap;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.ISpaceWithIdentity;
import org.w3c.dom.Document;
import java.util.*;

public class PaletteAccessor extends StandardAccessorImpl
{
	public PaletteAccessor()
	{	
	}
	
	public void onSource(INKFRequestContext aContext) throws Exception
	{	
		String entityId=aContext.getThisRequest().getArgumentValue("entity");
		IRepDeployedModules modules=(IRepDeployedModules)aContext.source("active:moduleList",IRepDeployedModules.class);
		NCODEPrototype endpoint=(NCODEPrototype)modules.componentForHash(entityId);
		if (endpoint==null)
		{	throw aContext.createFormattedException("EX_NCODE_RUNTIME", "MSG_NCODE_NO_INSTANCE", null, null, entityId);
		}
		NCODEConfig config=endpoint.getDataFlowConfig(aContext);

		AvailablePalettesRepresentation availablePalettes=aContext.source("active:availablePalettes",AvailablePalettesRepresentation.class);
		IHDSNodeList allPaletteSpaces=availablePalettes.getSpaces();
		
		//if override is posted use that otherwise look at save state for list of dynamic imports
		final IHDSNode dynamicImports;
		if (aContext.getThisRequest().argumentExists("imports"))
		{	dynamicImports=aContext.source("arg:imports",IHDSNode.class);
		}
		else
		{	IHDSNode hds=config.getRawConfig();
			dynamicImports=hds.getFirstNode("/cde/settings/dynamicImport__A");
		}
		
		final Set<ISpace> spaces=endpoint.findSpacesInScope(aContext,modules);
		
		IHDSPredicate f = new IHDSPredicate()
		{
			public boolean matches(IHDSNode aNode)
			{	String spaceId=aNode.getFirstValue("id").toString().trim();
				String spaceVersion=aNode.getFirstValue("version").toString().trim();
				boolean result=false;
				for (ISpace space: spaces)
				{	if (space instanceof ISpaceWithIdentity)
					{	ISpaceWithIdentity swi = (ISpaceWithIdentity)space;
						if (swi.getIdentifier().toString().equals(spaceId) && swi.getVersion().toString().equals(spaceVersion))
						{	result=true;
						}
					}
				}
				if (!result && dynamicImports!=null)
				{	for (IHDSNode diNode: dynamicImports.getChildren())
					{	String diId=(String)diNode.getFirstValue("id");
						if (diId.equals(spaceId))
						{	result=true;
							break;
							/*
							String diVersion=(String)diNode.getFirstValue("version");
							if (diVersion.equals(spaceVersion))
							{	result=true;
								break;
							}
							*/
						}
					}
				}
				return result;
			}
		};
		IHDSNodeList selectedPaletteSpaces=allPaletteSpaces.filter(f);

		PaletteRepresentation rep=new PaletteRepresentation();
		rep.addGroup("literals", "Literals", "", 0);
		rep.addGroup("core", "Builtin Accessors", "", 1);
		rep.addGroup("io", "Inputs/Outputs", "", 2);
		rep.addGroup("other", "Accessors", "", 3);
		rep.addGroup("exp", "Experimental", "", 4);
		availablePalettes.populatePalette(rep,selectedPaletteSpaces,modules,aContext);
		
		StaticPaletteItem spi;
		ValueValidatorFactory vvf=ValueValidatorFactory.getInstance();

		//find and add existing flows to palette
		
		ISpaceElements se=config.getElements(aContext);
		for (int i=0; i<se.size(); i++)
		{	String identifier=se.getIdentifier(i).toString();
			try
			{	LogicalEndpointPaletteItem lepi=new LogicalEndpointPaletteItem(identifier, null, 100, modules, aContext);
				lepi.setDynamic();
				lepi.setIcon("/tools/ncode/pub/icon.png");
				rep.addEndpoint("io",lepi,aContext);
			}
			catch (Exception e) {;}
		}
		
		
		spi=new StaticPaletteItem("df:comment", 0, "comment", null, null, null,null,null);
		spi.setValue(vvf.mNoValidation, "comment");
		spi.setCSSClass("comment");
		spi.setDocumentation("doc:lang:ncode:ui:comments");
		rep.addEndpoint("literals", spi,aContext);
		
		spi=new StaticPaletteItem("df:literal:absolute-identifier", 0, "Identifier", IIdentifier.class, null, null, null, null);
		spi.setValue(vvf.mURIValidation, "res:/");
		spi.setCSSClass("literal");
		spi.setDocumentation("doc:lang:ncode:ui:literals");
		rep.addEndpoint("literals", spi,aContext);
		
		spi=new StaticPaletteItem("df:literal:string", 0, "String", String.class, null, null,null,null);
		spi.setValue(vvf.mNoValidation, "String");
		spi.setCSSClass("literal");
		spi.setDocumentation("doc:lang:ncode:ui:literals");
		rep.addEndpoint("literals", spi,aContext);
		
		spi=new StaticPaletteItem("df:literal:boolean", 0, "Boolean", Boolean.class, null, null,null,null);
		spi.setValue(vvf.mBooleanValidation, "false");
		spi.setCSSClass("literal");
		spi.setDocumentation("doc:lang:ncode:ui:literals");
		rep.addEndpoint("literals", spi,aContext);
		
		spi=new StaticPaletteItem("df:literal:integer", 0, "Integer", Integer.class, null, null,null,null);
		spi.setValue(vvf.mIntegerValidation, "0");
		spi.setCSSClass("literal");
		spi.setDocumentation("doc:lang:ncode:ui:literals");
		rep.addEndpoint("literals", spi,aContext);
		
		spi=new StaticPaletteItem("df:literal:long", 0, "Long", Integer.class, null, null,null,null);
		spi.setValue(vvf.mIntegerValidation, "0");
		spi.setCSSClass("literal");
		spi.setDocumentation("doc:lang:ncode:ui:literals");
		rep.addEndpoint("literals", spi,aContext);
		
		spi=new StaticPaletteItem("df:literal:float", 0, "Float", Integer.class, null, null,null,null);
		spi.setValue(vvf.mFloatValidation, "0.0");
		spi.setCSSClass("literal");
		spi.setDocumentation("doc:lang:ncode:ui:literals");
		rep.addEndpoint("literals", spi,aContext);
		
		spi=new StaticPaletteItem("df:literal:double", 0, "Double", Integer.class, null, null,null,null);
		spi.setValue(vvf.mFloatValidation, "0.0");
		spi.setCSSClass("literal");
		spi.setDocumentation("doc:lang:ncode:ui:literals");
		rep.addEndpoint("literals", spi,aContext);
		

		spi=new StaticPaletteItem("df:literal:xml", 0, "XML", Document.class, null, null,null,null);
		spi.setValue(vvf.mXMLValidation, "<xml/>");
		spi.setCSSClass("literal");
		spi.setDocumentation("doc:lang:ncode:ui:literals");
		rep.addEndpoint("literals", spi,aContext);
		
		spi=new StaticPaletteItem("df:literal:hds", 0, "HDS", IHDSNode.class, null, null,null,null);
		spi.setValue(vvf.mHDSValidation, "<hds/>");
		spi.setCSSClass("literal");
		spi.setDocumentation("doc:lang:ncode:ui:literals");
		rep.addEndpoint("literals", spi,aContext);

		spi=new StaticPaletteItem("df:literal:pojo", 0, "Object", IHDSNode.class, null, null,null,null);
		spi.setValue(vvf.mPOJOValidation, "<literal type='java.lang.Object'/>");
		spi.setCSSClass("literal");
		spi.setDocumentation("doc:lang:ncode:ui:literals");
		rep.addEndpoint("literals", spi,aContext);

		
		
		spi=new StaticPaletteItem("df:response", 0, "Response", null, "response", Object.class,null,null);
		spi.setCSSClass("response");
		spi.setIcon("/tools/ncode/pub/nkrelief.png");
		rep.addEndpoint("io", spi,aContext);

		spi=new StaticPaletteItem("df:arg:arguments", 0, "arg: arguments", IHDSNode.class, null, null,null,null);
		spi.setCSSClass("argument");
		rep.addEndpoint("io", spi,aContext);
		
		spi=new StaticPaletteItem("df:arg:verb", 0, "arg:verb", String.class, null, null,null,null);
		spi.setCSSClass("argument");
		rep.addEndpoint("io", spi,aContext);

		spi=new StaticPaletteItem("df:arg:primary", 0, "arg:primary", String.class, null, null,null,null);
		spi.setCSSClass("argument");
		rep.addEndpoint("io", spi,aContext);

		spi=new StaticPaletteItem("df:dereference", 1, "String To Identifier", IIdentifier.class, "operand", String.class,null,null);
		spi.setDocumentation("doc:lang:ncode:ui:dereference");
		rep.addEndpoint("core", spi,aContext);
		
		
		//TODO refactor these into proper modules
		/*
		OrderedCheapMultiStringMap m=new OrderedCheapMultiStringMap(1);
		m.put("activeType", "xslt");
		LogicalEndpointPaletteItem lepi=new LogicalEndpointPaletteItem("apache.XSLTRuntime", m, 10, modules, aContext);
		lepi.overrideName("XSLT");
		rep.addEndpoint("other",lepi,aContext);
		
		rep.addEndpoint("other",new LogicalEndpointPaletteItem("GroovyRuntime", null, 0, modules, aContext),aContext);
		rep.addEndpoint("other",new LogicalEndpointPaletteItem("freemarker:runtime", null, 0, modules, aContext),aContext);
		//rep.addEndpoint("exp",new LogicalEndpointPaletteItem("layer1.DataScheme", null, 0, modules, aContext),aContext);
		*/
		
		aContext.createResponseFrom(rep);
	}
}
