package org.netkernel.lang.ncode;

import org.netkernel.lang.ncode.PaletteRepresentation.IPaletteItem;
import org.netkernel.lang.ncode.editor.IValueValidator;
import org.netkernel.layer0.meta.IEndpointArgumentMeta;
import org.netkernel.layer0.meta.IEndpointArgumentMeta.ArgumentType;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.representation.IBinaryStreamRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.util.OrderedCheapMultiStringMap;
import org.netkernel.urii.IIdentifier;

public class StaticPaletteItem implements IPaletteItem
{
	private final String mId;
	private final int mOrder;
	private final String mName;
	private final Class mResponseClass;
	private final String mArgName1;
	private final Class mArgClass1;
	private final String mArgName2;
	private final Class mArgClass2;
	private String mCSSClass;
	private IValueValidator mValueValidator;
	private String mInitialValue;
	private String mDocumentation;
	private String mDesc;
	private String mIcon;
	
	/* id,name,response-rep,input-rep,input-name
	 * 
	 * 
	 * 
	 * 
	 * 
	 */
	
	
	
	public StaticPaletteItem(String aId, int aOrder, String aName, Class aResponseClass, String aArgName1, Class aArgClass1, String aArgName2, Class aArgClass2)
	{	mId=aId;
		mOrder=aOrder;
		mName=aName;
		mResponseClass=aResponseClass;
		mArgName1=aArgName1;
		mArgClass1=aArgClass1;
		mArgName2=aArgName2;
		mArgClass2=aArgClass2;
	}
	
	public void setCSSClass(String aCSSClass)
	{	mCSSClass=aCSSClass;
	}
	
	public void setIcon(String aIcon)
	{	mIcon=aIcon;
	}

	public void setDesc(String aDesc)
	{	mDesc=aDesc;
	}

	public void setValue(IValueValidator aValueValidator, String aInitialValue)
	{
		mValueValidator=aValueValidator;
		mInitialValue=aInitialValue;
	}
	
	public void setDocumentation(String aDocumentation)
	{	mDocumentation=aDocumentation;
	}

	@Override
	public int compareTo(IPaletteItem o)
	{	return mOrder-o.getOrder();
	}
	
	public int getOrder()
	{	return mOrder;
	}
	
	public String getEndpointId()
	{	return mId;
	}
	
	public String getPaletteId()
	{	return mId;
	}

	public void serializeTo(HDSBuilder b)
	{	b.pushNode("endpoint");
		b.addNode("eid", mId);
		b.addNode("id", mId);
		b.addNode("name", mName);
		if (mDesc!=null)
		{	b.addNode("desc", mDesc);
		}
		if (mCSSClass!=null)
		{	b.addNode("cssClass",mCSSClass);
		}
		if (mIcon!=null)
		{	b.addNode("icon",mIcon);
		}
		if (mValueValidator!=null)
		{	b.addNode("hasValue",true);
			b.addNode("v",mInitialValue);
		}
		if (mResponseClass!=null)
		{	//b.pushNode("output__A");
			b.pushNode("output");
			b.pushNode("response");
			boolean isIdentifier=mResponseClass==IIdentifier.class;
			b.addNode("t", isIdentifier?IEndpointArgumentMeta.ArgumentType.IDENTIFIER.name():IEndpointArgumentMeta.ArgumentType.REPRESENTATION.name());
			b.addNode("s", isIdentifier?"":mResponseClass.getName());
			b.popNode();
			b.popNode();
			//b.popNode();
		}
		if (mArgName1!=null)
		{	b.addNode("valid", false);
			b.pushNode("input__A");
			b.pushNode("input");
			b.addNode("n", mArgName1);
			b.addNode("p", mArgName1);
			boolean isIdentifier=mArgClass1==IIdentifier.class;
			b.addNode("t", isIdentifier?IEndpointArgumentMeta.ArgumentType.IDENTIFIER.name():IEndpointArgumentMeta.ArgumentType.REPRESENTATION.name());
			b.addNode("s", isIdentifier?"":mArgClass1.getName());
			b.addNode("o", false);
			b.addNode("r", false);
			b.popNode();
			if (mArgName2!=null)
			{
				b.pushNode("input");
				b.addNode("n", mArgName2);
				b.addNode("p", mArgName2);
				isIdentifier=mArgClass2==IIdentifier.class;
				b.addNode("t", isIdentifier?IEndpointArgumentMeta.ArgumentType.IDENTIFIER.name():IEndpointArgumentMeta.ArgumentType.REPRESENTATION.name());
				b.addNode("s", isIdentifier?"":mArgClass2.getName());
				b.addNode("o", false);
				b.addNode("r", false);
				b.popNode();
			}
			b.popNode();		
		}
		else
		{	b.addNode("valid", true);
		}
		
		
		
		//b.importChildren(mDefn);

		b.popNode();
	}
	
	public IHDSNode validateInputs(IHDSNode aInputs, INKFRequestContext aContext)
	{
		HDSBuilder b=new HDSBuilder();
		if (mArgName1==null)
		{	b.addNode("valid", true);
		}
		else
		{	IHDSNode input=aInputs.getFirstNode("i");
			boolean connected=(Boolean)input.getFirstValue("c");
			b.addNode("valid", connected);
			
			if (mArgName1!=null)
			{	b.pushNode("input__A");
				b.pushNode("input");
				b.addNode("n", mArgName1);
				b.addNode("p", mArgName1);
				b.addNode("t", ArgumentType.REPRESENTATION.name());
				b.addNode("s", mArgClass1.getName());
				b.addNode("o", false);
				b.addNode("r", false);
				if (connected) b.addNode("i", 0);
				b.popNode();
				
				if (mArgName2!=null)
				{
					input=aInputs.getFirstNode("i[2]");
					connected=(Boolean)input.getFirstValue("c");
					b.pushNode("input");
					b.addNode("n", mArgName2);
					b.addNode("p", mArgName2);
					b.addNode("t", ArgumentType.REPRESENTATION.name());
					b.addNode("s", mArgClass2.getName());
					b.addNode("o", false);
					b.addNode("r", false);
					if (connected) b.addNode("i", 1);
					b.popNode();
				}
				
				b.popNode();			
			}			
		}
		
		
		return b.getRoot();
	}
	
	public IBinaryStreamRepresentation getIcon(INKFRequestContext aContext) throws Exception
	{	return null;
	}

	public IHDSNode validateValue(String aValue)
	{	if (mValueValidator!=null)
		{	return mValueValidator.validate(aValue);
		}
		else
		{	return IValueValidator.VALID;
		}
	}
	
	public OrderedCheapMultiStringMap getHardWiredInputs()
	{	return EndpointMetaWrapper.sNoHardArgs;
	}

	/*
	public boolean isIdentifierInput(String aInputProto)
	{	return true;
	}
	*/

	public ArgumentType getInputType(String aInputName)
	{	return ArgumentType.REPRESENTATION;
	}

	public String getDocumentation()
	{	return mDocumentation;
	}	
	
}
