package org.netkernel.lang.ncode;

import org.netkernel.ext.system.representation.IRepDeployedModules;
import org.netkernel.lang.dpml.representation.DPMLRepresentation;
import org.netkernel.lang.ncode.PaletteRepresentation.IPaletteItem;
import org.netkernel.layer0.meta.IEndpointArgumentMeta;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.impl.MetadataUtils;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.representation.IHDSIndexSpecification;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.impl.HDSIndex;
import org.netkernel.layer0.representation.impl.HDSPredicateFactory;
import org.netkernel.layer0.util.OrderedCheapMultiStringMap;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.urii.IMetaRepresentation;
import org.w3c.dom.Document;
import java.io.StringWriter;
import java.io.Writer;
import java.util.*;

public class ToDPMLAccessor extends StandardAccessorImpl
{
	private final Comparator<IHDSNode> mInputWireSorter;
	
	public ToDPMLAccessor()
	{
		mInputWireSorter = new Comparator<IHDSNode>()
		{	public int compare(IHDSNode arg0, IHDSNode arg1)
			{	String v0=arg0.getFirstValue("ec").toString();
				String v1=arg1.getFirstValue("ec").toString();
				return Integer.parseInt(v0)-Integer.parseInt(v1);
			}
		};
	}
	
	public void onSource(INKFRequestContext aContext) throws Exception
	{
		NCODEConfig config=aContext.source("arg:config",NCODEConfig.class);
		NCODEPrototype prototypeInstance=config.getOwningEndpoint();
		
		//IRepDeployedModules modules=(IRepDeployedModules)aContext.source("active:moduleList",IRepDeployedModules.class);
		//String entity=modules.hashForComponent(prototypeInstance);
		String entity=aContext.source("arg:entity",String.class);
		
		INKFRequest req=aContext.createRequest("active:nCoDEPalette");
		req.addArgument("entity", entity);
		PaletteRepresentation palette=(PaletteRepresentation)aContext.issueRequest(req);		
		
		StringWriter w=new StringWriter();
		XMLUtils.writeOpenTag(w, "closure");
		
		Map<String,Flow> flowsById=config.getFlowsById();
		for (Map.Entry<String, Flow> entry : flowsById.entrySet())
		{
			w.write("<closure ");
			XMLUtils.writeAttribute(w, "assignment", entry.getKey());
			w.write(">");

			final IHDSNode cn=entry.getValue().getConfig();
			String flowId=entry.getKey();
			IEndpointMeta flowMeta=(IEndpointMeta)config.getElementMeta(flowId);
			
			final IHDSNodeList endpoints=cn.getNodes("endpoint__A/endpoint");
			final IHDSNodeList wires=cn.getNodes("wire__A/wire");
			IHDSNode responseEndpoint=endpoints.filter(HDSPredicateFactory.namedChildStringEquals("p", "df:response")).getFirstNode();
			
			IHDSIndexSpecification eis = new IHDSIndexSpecification()
			{	public IHDSNodeList getNodesToIndex(IHDSNode aContextNode)
				{	return cn.getNodes("endpoint__A/endpoint");
				}
				public Object getKeyValue(IHDSNode aNodeToIndex)
				{	return aNodeToIndex.getFirstValue("i").toString();
				}
			};
			HDSIndex endpointsById=new HDSIndex(cn, eis);
			
			eis = new IHDSIndexSpecification()
			{	public IHDSNodeList getNodesToIndex(IHDSNode aContextNode)
				{	return cn.getNodes("wire__A/wire");
				}
				public Object getKeyValue(IHDSNode aNodeToIndex)
				{	return aNodeToIndex.getFirstValue("ee").toString();
				}
			};
			HDSIndex wiresByEnd=new HDSIndex(cn, eis);

			CountMap endpointReferenceCount=new CountMap();
			recurseFindMultipleReferences(responseEndpoint,endpointsById,wiresByEnd,endpointReferenceCount);
			
			for (Object id : endpointReferenceCount.keySet())
			{	int count=endpointReferenceCount.get(id);
				if (count>1)
				{	IHDSNode endpoint=endpointsById.get(id).getFirstNode();
					recurseWriteEndpoint(endpoint,endpointsById,wiresByEnd,endpointReferenceCount,w,(String)id,palette,IEndpointArgumentMeta.ArgumentType.REPRESENTATION,aContext,flowMeta);
				}
			}
			recurseWriteEndpoint(responseEndpoint,endpointsById,wiresByEnd,endpointReferenceCount,w,null,palette,IEndpointArgumentMeta.ArgumentType.REPRESENTATION,aContext,flowMeta);
			
			XMLUtils.writeCloseTag(w, "closure");
		}
		XMLUtils.writeCloseTag(w, "closure");
		
		String script=w.toString();
		/*
		// output dpml to console
		Document d=aContext.transrept(script, Document.class);
		org.netkernel.xml.util.XMLUtils.toXML(System.out, d, true, true,"UTF-8");
		*/
		DPMLRepresentation dpml=aContext.transrept(script, DPMLRepresentation.class);
		
		aContext.createResponseFrom(dpml);
	}
	
	private void recurseFindMultipleReferences(IHDSNode aNode, HDSIndex aEndpointsById, HDSIndex aWiresByEnd, CountMap aCounts) throws Exception
	{
		String id=aNode.getFirstValue("i").toString();
		int count=aCounts.increment(id);
		if (count>1) return;
		
		IHDSNodeList wiresIn=aWiresByEnd.get(id);
		for (IHDSNode wireIn : wiresIn)
		{	String wireStartId=wireIn.getFirstValue("se").toString();
			String wireStartCon=wireIn.getFirstValue("sc").toString();
			if (wireStartCon.equals("R"))
			{	IHDSNode wireStartEndpoint=aEndpointsById.get(wireStartId).getFirstNode();
				recurseFindMultipleReferences(wireStartEndpoint,aEndpointsById,aWiresByEnd,aCounts);
			}
		}
	}
	
	private final String ID_DF="df:";
	private final String ID_RESPONSE="df:response";
	private final String ID_DEREFERENCE="df:dereference";
	private final String ID_LITERAL="df:literal:";
	private final String ID_ARG="df:arg:";
	
	private void recurseWriteEndpoint(IHDSNode aNode, HDSIndex aEndpointsById, HDSIndex aWiresByEnd, CountMap aCounts, Writer w, String aAssignment, PaletteRepresentation aPalette, IEndpointArgumentMeta.ArgumentType aArgType, INKFRequestContext aContext, IEndpointMeta aFlowMeta) throws Exception
	{
		
		String proto=aNode.getFirstValue("p").toString();
		
		if (proto.startsWith(ID_DF))
		{	if (proto.equals(ID_RESPONSE))
			{	String id=aNode.getFirstValue("i").toString();
				IHDSNodeList wiresIn=aWiresByEnd.get(id);
				if (wiresIn.size()>0)
				{
					String wireStartId=wiresIn.getFirstValue("se").toString();
					IHDSNode wireStartEndpoint=aEndpointsById.get(wireStartId).getFirstNode();
					recurseWriteEndpoint(wireStartEndpoint,aEndpointsById,aWiresByEnd,aCounts,w,"response",aPalette,IEndpointArgumentMeta.ArgumentType.REPRESENTATION,aContext,aFlowMeta);
				}
			}
			else if (proto.equals(ID_DEREFERENCE))
			{	w.write("<dereference ");
				if (aAssignment!=null)
				{	XMLUtils.writeAttribute(w, "assignment", aAssignment);
				}
				w.write(">");
				String id=aNode.getFirstValue("i").toString();
				String wireStartId=aWiresByEnd.get(id).getFirstValue("se").toString();
				IHDSNode wireStartEndpoint=aEndpointsById.get(wireStartId).getFirstNode();
				recurseWriteEndpoint(wireStartEndpoint,aEndpointsById,aWiresByEnd,aCounts,w,null,aPalette,IEndpointArgumentMeta.ArgumentType.REPRESENTATION,aContext,aFlowMeta);
				XMLUtils.writeCloseTag(w, "dereference");
			}
			else if (proto.startsWith(ID_LITERAL))
			{	String type=proto.substring(ID_LITERAL.length());
				String value=(String)aNode.getFirstValue("v");
				boolean identifierLiteral=type.equals("identifier") || type.equals("absolute-identifier");
				boolean deref=(aArgType==IEndpointArgumentMeta.ArgumentType.STRING || aArgType==IEndpointArgumentMeta.ArgumentType.IDENTIFIER) && !identifierLiteral;
				if (type.equals("pojo"))
				{	w.write(value);
				}
				else
				{	if (deref)
					{	w.write("<dereference>");
					}
					w.write("<literal ");
					if (aAssignment!=null)
					{	XMLUtils.writeAttribute(w, "assignment", aAssignment);
					}
					XMLUtils.writeAttribute(w, "type", type);
					w.write(">");
					if (type.equals("xml") || type.equals("hds"))
					{
						w.write(value);
					}
					else
					{	if (value!=null)
						{	w.write(XMLUtils.escape(value));
						}
					}
					XMLUtils.writeCloseTag(w, "literal");
					if (deref)
					{	w.write("</dereference>");
					}
				}
			}
			else if (proto.startsWith(ID_ARG))
			{	String argName=proto.substring(ID_ARG.length());
				String identifier="arg:"+argName;
				boolean deref=(aArgType==IEndpointArgumentMeta.ArgumentType.STRING || aArgType==IEndpointArgumentMeta.ArgumentType.IDENTIFIER);
				if (deref)
				{	w.write("<dereference>");
				}
				w.write("<literal type=\"identifier\" ");
				if (aAssignment!=null)
				{	XMLUtils.writeAttribute(w, "assignment", aAssignment);
				}
				w.write(">");
				w.write(identifier);
				w.write("</literal>");
				if (deref)
				{	w.write("</dereference>");
				}
			}
			else
			{	writeRequest(aNode,aEndpointsById,aWiresByEnd,aCounts,w,aAssignment,aPalette,aArgType,aContext,aFlowMeta);
			}
		}
		else
		{	writeRequest(aNode,aEndpointsById,aWiresByEnd,aCounts,w,aAssignment,aPalette,aArgType,aContext,aFlowMeta);
		}
		
	}

	private void writeRequest(IHDSNode aNode, HDSIndex aEndpointsById, HDSIndex aWiresByEnd, CountMap aCounts, Writer w, String aAssignment, PaletteRepresentation aPalette, IEndpointArgumentMeta.ArgumentType aArgType, INKFRequestContext aContext, IEndpointMeta aFlowMeta) throws Exception
	{
		boolean autoDeref=aArgType==IEndpointArgumentMeta.ArgumentType.STRING || aArgType==IEndpointArgumentMeta.ArgumentType.IDENTIFIER;
		if (autoDeref)
		{	w.write("<dereference>");
		}
		w.write("<request ");
		if (aAssignment!=null)
		{	XMLUtils.writeAttribute(w, "assignment", aAssignment);
		}
		w.write(">");
		
		String id=aNode.getFirstValue("i").toString();
		String proto=aNode.getFirstValue("p").toString();
		
		IPaletteItem pi=null;
		try
		{	pi=aPalette.getPaletteItem(proto,aContext);
		}
		catch (Exception e)
		{	//unresolvable endpoint - this will fail at execution time
			XMLUtils.write(w, "identifier", "meta:"+proto);
		}
		
		if (pi!=null)
		{
			String identifier="meta:"+pi.getEndpointId();
			XMLUtils.write(w, "identifier", identifier);
	
			OrderedCheapMultiStringMap hwi=pi.getHardWiredInputs();
			for (int i=0; i<hwi.size(); i++)
			{	String value=hwi.getValue(i);
				if (value!=null)
				{	w.write("<argument ");
					XMLUtils.writeAttribute(w, "name", hwi.getKey(i));
					w.write(">");	
					w.write("<literal type='absolute-identifier'>");
					w.write(XMLUtils.escape(value));
					XMLUtils.writeCloseTag(w, "literal");
					XMLUtils.writeCloseTag(w, "argument");
				}
			}
	
			IHDSNodeList wiresIn=aWiresByEnd.get(id).sort(mInputWireSorter);
			IHDSNodeList inputs=null;
			if (wiresIn.size()>0)
			{	inputs=aNode.getNodes("input__A/input");
			}
			for (IHDSNode wireIn : wiresIn)
			{	
				int i=Integer.parseInt(wireIn.getFirstValue("ec").toString());
				IHDSNode inputNode=inputs.get(i);
				String inputName=(String)inputNode.getFirstValue("n");
				String inputProto=(String)inputNode.getFirstValue("p");
				w.write("<argument ");
				XMLUtils.writeAttribute(w, "name", inputName);
				w.write(">");
				
				String wireStartId=wireIn.getFirstValue("se").toString();
				String wireStartCon=wireIn.getFirstValue("sc").toString();
				if (wireStartCon.equals("R"))
				{
					int count=aCounts.get(wireStartId);
					if (count==1)
					{	IHDSNode wireStartEndpoint=aEndpointsById.get(wireStartId).getFirstNode();
						recurseWriteEndpoint(wireStartEndpoint,aEndpointsById,aWiresByEnd,aCounts,w,null,aPalette,pi.getInputType(inputProto),aContext,aFlowMeta);
					}
					else
					{	
						IEndpointArgumentMeta.ArgumentType type=pi.getInputType(inputProto);
						String reference="this:"+wireStartId;
						if (type==IEndpointArgumentMeta.ArgumentType.STRING)
						{	w.write("<dereference>");
							w.write(reference);
							w.write("</dereference>");
						}
						else
						{	w.write(reference);
						}
					}
				}
				else
				{	//reference to injected identifier
					IHDSNode wireStartEndpoint=aEndpointsById.get(wireStartId).getFirstNode();
					int iid=-Integer.parseInt(wireStartCon);
					String injectedIdentifier=(String)wireStartEndpoint.getFirstValue("output/injected__A/injected["+iid+"]/uri");
					w.write(injectedIdentifier);
				}
				
				XMLUtils.writeCloseTag(w, "argument");
			}
		}
		
		XMLUtils.writeCloseTag(w, "request");
		if (autoDeref)
		{	w.write("</dereference>");
		}
	}	
}