package org.netkernel.lang.ncode;

import java.util.Iterator;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.netkernel.layer0.meta.IEndpointArgumentMeta;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.impl.MetadataUtils;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;

public class Utils
{
	public static final String ARRAY_INDICATOR_SUFFIX="__A";
	
	public static IHDSNode JSONToHDS(JSONObject aJSON) throws JSONException
	{
		HDSBuilder b=new HDSBuilder();
		recurseJSONToHDS(b,aJSON);
		return b.getRoot();
	}
	
	public static JSONObject HDSToJSON(IHDSNode  aHDS) throws JSONException
	{
		JSONObject result;
		Object o=recurseHDSToJSON(aHDS);
		if (o instanceof JSONObject)
		{	result=(JSONObject)o;
		}
		else
		{	result = new JSONObject();
			result.put(aHDS.getName(), o);
		}
		return result;
	}
	
	private static Object recurseHDSToJSON(IHDSNode  aHDS) throws JSONException
	{
		Object result=null;
		
		String name=aHDS.getName();
		Object value=aHDS.getValue();
		if (value==null)
		{	if (name!=null && name.endsWith(ARRAY_INDICATOR_SUFFIX))
			{	JSONArray a = new JSONArray();
				for (IHDSNode n : aHDS.getChildren())
				{	Object c=recurseHDSToJSON(n);
					if (c!=null) a.put(c);
				}
				result=a;
			}
			
			else if (aHDS.getChildren().length>0)
			{	JSONObject o = new JSONObject();
				for (IHDSNode n : aHDS.getChildren())
				{	Object c=recurseHDSToJSON(n);
					if (c!=null)
					{	String cname=n.getName();
						if (cname.endsWith(ARRAY_INDICATOR_SUFFIX))
						{	cname=cname.substring(0,cname.length()-ARRAY_INDICATOR_SUFFIX.length());
						}
						o.put(cname, c);
					}
				}
				result=o;
			}
			else
			{	result=JSONObject.NULL;
			}
		}
		else if (value instanceof String || value instanceof Boolean || value instanceof Number)
		{	result=value;
		}
		return result;
	}
	
	
	
	private static void recurseJSONToHDS(HDSBuilder b, JSONObject j) throws JSONException
	{	for (Iterator<String> i=j.keys(); i.hasNext(); )
		{	String key=i.next();
			Object object=j.get(key);
			writeJSONToHDS(b,key,object);
		}
	}
	
	private static void recurseJSONArrayToHDS(HDSBuilder b, String name, JSONArray j) throws JSONException
	{	int l=j.length();
		for (int i=0; i<l; i++)
		{	Object object=j.get(i);
			writeJSONToHDS(b,name,object);
		}
	}
	
	private static void writeJSONToHDS(HDSBuilder b,String key, Object object) throws JSONException
	{	if (object instanceof JSONObject)
		{	b.pushNode(key);
			recurseJSONToHDS(b,(JSONObject)object);
			b.popNode();
		}
		else if (object instanceof JSONArray)
		{	b.pushNode(key+ARRAY_INDICATOR_SUFFIX);
			recurseJSONArrayToHDS(b,key,(JSONArray)object);
			b.popNode();
		}
		else if (object instanceof Number || object instanceof String || object instanceof Boolean)
		{	b.addNode(key,object);
		}
		else if (object==JSONObject.NULL)
		{	b.addNode(key,null);
		}
	}
	
	public static boolean shouldConvertArgumentToString(IEndpointMeta aEndpointMeta, String aArgumentName)
	{	IEndpointArgumentMeta argMeta=MetadataUtils.getEndpointIdentifierPart(aEndpointMeta, aArgumentName);
		boolean asString=argMeta!=null && argMeta.getArgumentType()==IEndpointArgumentMeta.ArgumentType.STRING;
		return asString;
	}
	
	
}
