package org.netkernel.lang.ncode.builtin;

import org.netkernel.layer0.meta.impl.SourcedArgumentMetaImpl;
import org.netkernel.layer0.nkf.*;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

public class ConditionalLoopAccessor extends StandardAccessorImpl {

  public ConditionalLoopAccessor() {
    declareThreadSafe();
    declareArgument(new SourcedArgumentMetaImpl("condition","condition to control the loop",null,new Class[] { Boolean.class}));
    declareArgument(new SourcedArgumentMetaImpl("do","service to call for each iteration",null,null));
    declareArgument(new SourcedArgumentMetaImpl("tolerant","true if we should ignore failed do's","false",new Class[] { Boolean.class}));
    declareSourceRepresentation(Iterable.class);
  }

  public void onSource(INKFRequestContext aContext) throws Exception {
    // Determine mode of operation
    INKFRequestReadOnly thisReq=aContext.getThisRequest();

    boolean tolerant = thisReq.argumentExists("tolerant") && aContext.source("arg:tolerant",Boolean.class);

    boolean testBefore = Boolean.FALSE;
    if (thisReq.getIdentifier().startsWith("active:ncodeWhile")){
      testBefore = Boolean.TRUE;
    }

    // PRE loop condition check
    INKFResponseReadOnly conditionResp;
    Boolean condition;
    INKFRequest conditionReq;

    if (testBefore){
      conditionReq=aContext.createRequest("arg:condition");
      conditionReq.setRepresentationClass(Boolean.class);
      conditionResp=aContext.issueRequestForResponse(conditionReq);
      condition=(Boolean)conditionResp.getRepresentation();
    }
    else{
      condition=Boolean.FALSE;
    }

    // Provide for the do responses
    List responseList = new ArrayList();

    // Loop
    // iteration WHILE the condition is TRUE (possibly no do-request at all)
    // or UNTIL the condition is TRUE (at least one do-request is executed)
    while(condition == testBefore){
      INKFRequest doReq=aContext.createRequest("arg:do");
      Object respIn;
      try {
        respIn=aContext.issueRequestForResponse(doReq);
      }
      catch (NKFException e) {
        if (tolerant) {
          respIn=e;
        }
        else {
          throw e;
        }
      }
      responseList.add(respIn);

      // in loop condition check
      conditionReq=aContext.createRequest("arg:condition");
      conditionReq.setRepresentationClass(Boolean.class);
      conditionResp=aContext.issueRequestForResponse(conditionReq);
      condition=(Boolean)conditionResp.getRepresentation();
    }

    //Create Representation List as response
    ArrayList representationList=new ArrayList(responseList.size());
    for(Object resp: responseList){
      Object rep;
      if(resp instanceof INKFResponseReadOnly) {
        INKFResponseReadOnly r=(INKFResponseReadOnly)resp;
        rep=r.getRepresentation();
      }
      else {
        rep=resp;  //Exception
      }
      representationList.add(rep);
    }
    aContext.createResponseFrom(Collections.unmodifiableList(representationList));
  }
}
