package org.netkernel.lang.ncode.builtin;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.netkernel.container.IKernel;
import org.netkernel.layer0.meta.impl.IdentifierArgumentMetaImpl;
import org.netkernel.layer0.meta.impl.SourcedArgumentMetaImpl;
import org.netkernel.layer0.nkf.INKFAsyncRequestHandle;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.urii.ParsedIdentifierImpl;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.layer0.urii.ValueSpace;
import org.netkernel.layer0.util.RequestBuilder;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.urii.IIdentifier;

public class ForEachAccessor extends StandardAccessorImpl
{
	public ForEachAccessor()
	{	declareThreadSafe();
		declareArgument(new SourcedArgumentMetaImpl("collection","collection of items",null,new Class[] {Iterable.class}));
		declareArgument(new SourcedArgumentMetaImpl("do","service to call for each item",null,null));
		declareArgument(new SourcedArgumentMetaImpl("async","if >=0 then request do's asynchronously with this arg as a timeout in millis","-1",new Class[] { Long.class}));
		declareArgument(new SourcedArgumentMetaImpl("tolerant","true if we should ignore failed do's","false",new Class[] { Boolean.class}));
		declareSourceRepresentation(Iterable.class);
	}
	
	public void onSource(INKFRequestContext aContext) throws Exception
	{
		//determine mode of operation
		INKFRequestReadOnly thisReq=aContext.getThisRequest();
		long timeout=-1;
		if (thisReq.argumentExists("async"))
		{	timeout=aContext.source("arg:async",Long.class);
		}
		boolean tolerant=thisReq.argumentExists("tolerant") && aContext.source("arg:tolerant",Boolean.class);
		
		//get collection
		Iterable operand=aContext.source("arg:collection",Iterable.class);
		List operandList = new ArrayList();
		for (Object item: operand)
		{	operandList.add(item);
		}
		
		//issue the requests
		int count=operandList.size();
		List responseList = new ArrayList(operandList.size());
		IKernel k=aContext.getKernelContext().getKernel();
		for (int index=0; index<count; index++)
		{
			INKFRequest doReq=aContext.createRequest("arg:do");
			ValueSpace values=new ValueSpace(3);
			values.put("foreach:item", operandList.get(index), null);
			values.put("foreach:count", count, null);
			values.put("foreach:index", index, null);
			values.onCommissionSpace(k);
			doReq.injectDurableRequestScope(values);
			
			if (timeout>=0)
			{	INKFAsyncRequestHandle handle=aContext.issueAsyncRequest(doReq);
				responseList.add(handle);
			}
			else
			{	Object respIn;
				try
				{	respIn=aContext.issueRequestForResponse(doReq);
				}
				catch (NKFException e)
				{	if (tolerant)
					{	respIn=e;
					}
					else
					{	throw e;
					}
				}
				responseList.add(respIn);
			}
		}
		
		//join async requests if necessary
		if (timeout>=0)
		{	for (int i=0; i<responseList.size(); i++)
			{	INKFAsyncRequestHandle handle=(INKFAsyncRequestHandle)responseList.get(i);
				Object respIn;
				try
				{	respIn=handle.joinForResponse(timeout);
					if(respIn==null)
					{	//Timed Out
						if(!tolerant)
						{	throw new NKFException("forEach Timeout", "For operand item number "+i+" timed out after "+timeout);							
						}
					}
				}
				catch (NKFException e)
				{	if (tolerant)
					{	respIn=e;
					}
					else
					{	throw e;
					}
				}
				responseList.set(i,respIn);
			}
		}
		
		//Create Representation List as response
		ArrayList representationList=new ArrayList(responseList.size());
		for(Object resp: responseList)
		{	Object rep;
			if(resp instanceof INKFResponseReadOnly)
			{	INKFResponseReadOnly r=(INKFResponseReadOnly)resp;
				rep=r.getRepresentation();
			}
			else
			{	rep=resp;	//Must be exception or null from timeout
			}
			representationList.add(rep);
		}
		aContext.createResponseFrom(Collections.unmodifiableList(representationList));
		
	}
}