package org.netkernel.lang.ncode.editor;

import org.json.JSONObject;
import org.netkernel.container.IKernel;
import org.netkernel.container.IMessages;
import org.netkernel.ext.system.representation.IRepDeployedModules;
import org.netkernel.ext.system.representation.impl.ProtoMap;
import org.netkernel.lang.ncode.AvailablePalettesRepresentation;
import org.netkernel.lang.ncode.NCODEPrototype;
import org.netkernel.lang.ncode.EndpointMetaWrapper;
import org.netkernel.lang.ncode.PaletteRepresentation;
import org.netkernel.lang.ncode.PaletteRepresentation.IPaletteItem;
import org.netkernel.lang.ncode.Utils;
import org.netkernel.layer0.bnf.GrammarParseException;
import org.netkernel.layer0.boot.IModule;
import org.netkernel.layer0.boot.ModuleManager;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.IModuleMeta;
import org.netkernel.layer0.meta.IPrototypeMeta;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.meta.impl.EndpointMetaBuilder;
import org.netkernel.layer0.meta.impl.MetadataUtils;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.representation.IBinaryStreamRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.representation.impl.HDSPredicateFactory;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.layer0.util.PairList;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.StandardModule;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.impl.Version;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import java.util.*;

public class NCODEEditor extends StandardAccessorImpl
{
	public NCODEEditor()
	{	this.declareThreadSafe();
	}
	
	public void onSource(INKFRequestContext aContext) throws Exception
	{
		//Thread.sleep(2000);
		String entity=aContext.getThisRequest().getArgumentValue("entity");
		String type=aContext.getThisRequest().getArgumentValue("type");
		if (type.equals("edit"))
		{	onWrapper(aContext,entity);
		}
		else if (type.equals("editor"))
		{	onTemplate(aContext,entity);
		}
		else if (type.equals("load"))
		{	onLoad(aContext,entity);
		}
		else if (type.equals("save"))
		{	onSave(aContext,entity);
		}
		else if (type.equals("palette"))
		{	onPalette(aContext,entity);
		}
		else if (type.equals("validateGrammar"))
		{	onValidateGrammar(aContext);
		}
		else if (type.equals("validateArguments"))
		{	onValidateArguments(aContext,entity);
		}
		else if (type.equals("validateValue"))
		{	onValidateValue(aContext,entity);
		}
		else if (type.equals("availablePalette"))
		{	onAvailablePalette(aContext,entity);
		}
		else if (type.equals("doc"))
		{	onDoc(aContext,entity);
		}
		else if (type.equals("allInstances"))
		{	onAllInstances(aContext);
		}
		else if (type.equals("redirect"))
		{	onRedirect(aContext,entity);
		}
	}
	
	private void throwNoInstanceException(INKFRequestContext aContext, String aEntity) throws Exception
	{	throw aContext.createFormattedException("EX_NCODE_RUNTIME", "MSG_NCODE_NO_INSTANCE", null, null, aEntity);
	}
	
	public void onRedirect(INKFRequestContext aContext, String aEntity) throws Exception
	{	int i=aEntity.indexOf("/");
		String moduleId=aEntity.substring(0,i);
		String instanceId=aEntity.substring(i+1);
		/*
		IModule foundModule=null;
		List<IModule> modules=ModuleManager.getSingleton().getModules();
		for (IModule module: modules)
		{	if (module.getMeta().getIdentifier().equals(moduleId))
			{	foundModule=module;
			}
		}
		if (foundModule==null || !(foundModule instanceof StandardModule)) return;
		StandardModule sm=(StandardModule)foundModule;
		*/
		
		IRepDeployedModules dm=(IRepDeployedModules)aContext.source("active:moduleList",IRepDeployedModules.class);
		IHDSNodeList spaceNodes=dm.getHierarchy().getNodes("/modules/module").filter(HDSPredicateFactory.namedChildStringEquals("id", moduleId)).getNodes("spaces//space");
		for (IHDSNode spaceNode: spaceNodes)
		{
			String spaceHash=spaceNode.getFirstValue("hash").toString();
			ISpaceMeta spaceMeta=(ISpaceMeta)dm.metadataForHash(spaceHash);
			ISpaceElements els=spaceMeta.getElements();
			for (i=0; i<els.size(); i++)
			{
				IEndpoint ep=els.getPhysicalEndpoint(i);
				//System.out.println("ep: "+ep);
				if (ep instanceof NCODEPrototype)
				{	String name=ep.toString();
					if (name.equals(instanceId))
					{	String epHash=dm.hashForComponent(ep);
						String url="/tools/ncode/edit/"+epHash;
						aContext.createResponseFrom("").setHeader("httpResponse:/redirect", url);
						return;
					}
				}
			}
		}
		
	}
	
	
	public void onAllInstances(INKFRequestContext aContext) throws Exception
	{	ProtoMap protoMap=aContext.source("active:moduleListProtoMap",ProtoMap.class);
		String protoHash=protoMap.getPrototypeHashFor(NCODEPrototype.class);
		String url="/tools/ae/view/prototypeReferences/"+protoHash;
		aContext.sink("httpResponse:/redirect", url);
	}
		
	
	public void onDoc(INKFRequestContext aContext, String aEntity) throws Exception
	{
		String endpointId=aContext.source("httpRequest:/param/eid", String.class);
		if (endpointId!=null)
		{	IRepDeployedModules modules=(IRepDeployedModules)aContext.source("active:moduleList",IRepDeployedModules.class);
			String epHash=modules.hashForSpaceElement(endpointId);
			if (epHash==null)
			{	String docId=null;
				try
				{	INKFRequest req=aContext.createRequest("active:nCoDEPalette");
					req.addArgument("entity", aEntity);
					PaletteRepresentation rep=(PaletteRepresentation)aContext.issueRequest(req);
					IPaletteItem item=rep.getPaletteItem(endpointId, aContext);
					docId=item.getDocumentation();
				}
				catch (Exception e) {;}
				if (docId==null)
				{	String doc="<div>No documentation found.</div>";
					aContext.createResponseFrom(doc);
				}
				else
				{	String doc=renderDoc(docId, aContext);
					String name=getDocTitle(docId,aContext);
					aContext.createResponseFrom(createDocResponse(name,doc,false,aContext));
				}
			}
			else
			{	
				String name="";
				INKFRequest req3=aContext.createRequest("active:nkseSearch");
				req3.addArgumentByValue("operand","eid:\""+endpointId+"\"");
				req3.addArgumentByValue("analyzer","org.apache.lucene.analysis.WhitespaceAnalyzer");
				req3.setRepresentationClass(IHDSNode.class);
				IHDSNode searchResult=(IHDSNode)aContext.issueRequest(req3);
				String doc=null;
				if (searchResult!=null)
				{	IHDSNodeList fields=searchResult.getNodes("/hits/hit/field");
					String documentID=(String)fields.filter(HDSPredicateFactory.namedChildStringEquals("name", "documentID")).getFirstValue("value");
					if (documentID!=null)
					{	//use existing documentation
						doc=renderDoc(documentID, aContext);
						name=getDocTitle(documentID,aContext);
					}
				}
				boolean warning=doc==null;
				if (warning)
				{	//generate a dynamic reference doc
					String docsrc="{endpoint}"+endpointId+"{/endpoint}";
					req3=aContext.createRequest("active:rawDocRenderer");
					req3.addArgumentByValue("docsrc",docsrc);
					req3.setRepresentationClass(String.class);
					doc=(String)aContext.issueRequest(req3);
					IMetaRepresentation epMeta=modules.metadataForHash(epHash);
					name=epMeta.getName();
				}
				aContext.createResponseFrom(createDocResponse(name, doc, warning, aContext));
			}
		}
		else
		{	String docId=aContext.source("httpRequest:/param/did", String.class);
			String name=getDocTitle(docId,aContext);
			String doc=renderDoc(docId, aContext);
			aContext.createResponseFrom(createDocResponse(name, doc, false, aContext));
		}
		
	}
	
	private String getDocTitle(String aDocId, INKFRequestContext aContext) throws Exception
	{	INKFRequest req=aContext.createRequest("active:documentationEntry");
		req.addArgument("id",aDocId);
		req.setRepresentationClass(IHDSNode.class);
		IHDSNode docEntry=(IHDSNode)aContext.issueRequest(req);
		String title= (String)docEntry.getFirstValue("/doc/title");
		return title;
	}
	
	private String createDocResponse(String aName, String aDoc, boolean aWarning, INKFRequestContext aContext) throws Exception
	{	StringBuilder sb=new StringBuilder(512);
		sb.append("<div ep-name='");
		sb.append(aName);
		sb.append("'>");
		sb.append(aDoc);
		if (aWarning)
		{	sb.append("<p>No documentation found for this endpoint. ");
			sb.append("This may either be because no documentation exists or because the search index is out of date. ");
			sb.append("Consider <a href='/panel/urn:org:netkernel:nkse:control:panel:tools'>rebuilding the search index</a>.</p>");
		}
		sb.append("</div>");
		return sb.toString();
	}

	private String renderDoc(String aDocId, INKFRequestContext aContext) throws Exception
	{	INKFRequest req3=aContext.createRequest("active:documentationEntry");
		req3.addArgument("id",aDocId);
		Object docentry=aContext.issueRequest(req3);
		req3=aContext.createRequest("active:docRendererInner");
		req3.addArgumentByValue("docentry",docentry);
		req3.setRepresentationClass(String.class);
		return (String)aContext.issueRequest(req3);
	}

	
	public void onWrapper(INKFRequestContext aContext, String aEntity) throws Exception
	{	INKFRequest req=aContext.createRequest("active:xrl2");
		req.addArgument("template", "res:/org/netkernel/lang/ncode/editor/wrapper.xml");
		req.addArgument("content", "res:/tools/ncode/editor/"+aEntity);
		INKFResponseReadOnly respIn=aContext.issueRequestForResponse(req);
		INKFResponse respOut=aContext.createResponseFrom(respIn);
		respOut.setMimeType("text/html");
	}
		
	public void onTemplate(INKFRequestContext aContext, String aEntity) throws Exception
	{	IRepDeployedModules modules=(IRepDeployedModules)aContext.source("active:moduleList",IRepDeployedModules.class);
		if (modules.metadataForHash(aEntity)==null)
		{	throwNoInstanceException(aContext,aEntity);
		}
		int i=aEntity.indexOf("/");
		String spaceHash=aEntity.substring(0,i);		
		IModuleMeta meta=(IModuleMeta)modules.metadataForHash(spaceHash);
		String moduleId=meta.getIdentifier();
		i=moduleId.lastIndexOf(':');
		i=moduleId.lastIndexOf(':', i-1);
		moduleId=moduleId.substring(i+1);
		
		String template=aContext.source("res:/org/netkernel/lang/ncode/editor/template.xml",String.class);
		INKFRequest req = aContext.createRequest("active:freemarker");
		req.addArgumentByValue("operator", template);
		req.addArgumentByValue("entityId",aEntity);
		req.addArgumentByValue("idPrefix",moduleId+":");
		template=(String)aContext.issueRequest(req);
		aContext.createResponseFrom(template);
	}
	
	public void onPalette(INKFRequestContext aContext, String aEntity) throws Exception
	{	INKFRequest req=aContext.createRequest("active:nCoDEPalette");
		req.addArgument("entity", aEntity);
		JSONObject json=aContext.source("httpRequest:/param/dynamicImports",JSONObject.class);
		if (json!=null)
		{	req.addArgumentByValue("imports",Utils.JSONToHDS(json).getFirstNode("*"));
		}
		PaletteRepresentation rep=(PaletteRepresentation)aContext.issueRequest(req);
		IHDSNode hds=rep.getPaletteAsHDS();
		json=Utils.HDSToJSON(hds);
		aContext.createResponseFrom(json.toString());
	}
	
	public void onLoad(INKFRequestContext aContext, String aEntity) throws Exception
	{
		IRepDeployedModules modules=(IRepDeployedModules)aContext.source("active:moduleList",IRepDeployedModules.class);
		NCODEPrototype endpoint=(NCODEPrototype)modules.componentForHash(aEntity);
		if (endpoint==null)
		{	throwNoInstanceException(aContext,aEntity);
		}
		IHDSNode hds;
		Object loaded=endpoint.load();
		if (loaded instanceof IHDSNode)
		{	hds=(IHDSNode)loaded;
		}
		else
		{	hds=aContext.transrept(loaded,IHDSNode.class);
		}
		HDSBuilder b=new HDSBuilder();
		b.importChildren(hds);
		//generate input definitions for each script and insert into data
		for (IHDSNode flow : b.getRoot().getNodes("/cde/flow__A/flow"))
		{
			String grammar=(String)flow.getFirstValue("grammar");
			try
			{	Document grammarDoc=aContext.transrept(grammar, Document.class);
				IHDSNode args=getInitialArgsForGrammar(grammarDoc, aContext);
				b.setCursor(flow);
				b.importChildren(args);
			}
			catch (Exception e)
			{	e.printStackTrace();
			}
		}
		//add operator verb support field
		b.setCursor(b.getRoot().getFirstNode("*"));
		b.addNode("operatorVerbs", endpoint.getOperatorVerbSupport());

		
		hds=b.getRoot();
		JSONObject json=Utils.HDSToJSON(hds);
		INKFResponse resp=aContext.createResponseFrom(json.toString());
		resp.setMimeType("application/json");		
	}
	
	
	public void onSave(INKFRequestContext aContext, String aEntity) throws Exception
	{	
		JSONObject json=aContext.source("httpRequest:/param/state",JSONObject.class);
		IHDSNode hds=Utils.JSONToHDS(json);
		IBinaryStreamRepresentation bs=aContext.transrept(hds, IBinaryStreamRepresentation.class);

		IRepDeployedModules modules=(IRepDeployedModules)aContext.source("active:moduleList",IRepDeployedModules.class);
		NCODEPrototype endpoint=(NCODEPrototype)modules.componentForHash(aEntity);
		if (endpoint==null)
		{	throwNoInstanceException(aContext,aEntity);
		}
		endpoint.save(bs);
		
		aContext.createResponseFrom("ok");
	}	
	
	public void onValidateGrammar(INKFRequestContext aContext) throws Exception
	{
		HDSBuilder b = new HDSBuilder();
		try
		{	Document grammarDoc =aContext.source("httpRequest:/param/grammar", Document.class);
			try
			{	IHDSNode args=getInitialArgsForGrammar(grammarDoc,aContext);
				b.importChildren(args);
				b.addNode("valid", "");
			}
			catch (GrammarParseException e)
			{	
				StringBuilder sb=new StringBuilder();
				PairList pl=e.getValidationErrors();
				for (int i=0; i<pl.size(); i++)
				{
					Node loc=(Node)pl.getValue1(i);
					sb.append(XMLUtils.getPathFor(loc));
					sb.append(": ");
					String msg=(String)pl.getValue2(i);
					sb.append(msg);
					sb.append(" ");
				}
				b.addNode("valid", sb.toString());
			}
			catch(Exception e)
			{	b.addNode("valid", e.getMessage());
			}
		}
		catch (Exception e)
		{	b.addNode("valid", "Malformed XML");
		}
		
		JSONObject jsonResult=Utils.HDSToJSON(b.getRoot());	
		INKFResponse resp=aContext.createResponseFrom(jsonResult.toString());
		resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
	}
	
	private IHDSNode getInitialArgsForGrammar(Document aGrammar, INKFRequestContext aContext) throws Exception
	{	IMessages messages=aContext.getKernelContext().getKernel().getLogger();
		EndpointMetaBuilder epm=new EndpointMetaBuilder();
		MetadataUtils.parseGrammar(aGrammar.getDocumentElement(), epm, messages);
		IEndpointMeta meta=epm.getMeta(messages);
		EndpointMetaWrapper emw=new EndpointMetaWrapper(meta, null, aContext);
		IHDSNode args=emw.getInitialArgs();
		return args;
	}
	
	public void onValidateArguments(INKFRequestContext aContext, String aEntity) throws Exception
	{	
		JSONObject arguments=aContext.source("httpRequest:/param/arguments", JSONObject.class);
		IHDSNode n=Utils.JSONToHDS(arguments);
		
		INKFRequest req=aContext.createRequest("active:nCoDEPalette");
		req.addArgument("entity", aEntity);
		JSONObject json=aContext.source("httpRequest:/param/dynamicImports",JSONObject.class);
		if (json!=null)
		{	req.addArgumentByValue("imports",Utils.JSONToHDS(json).getFirstNode("*"));
		}
		PaletteRepresentation rep=(PaletteRepresentation)aContext.issueRequest(req);
		
		HDSBuilder b=new HDSBuilder();
		b.pushNode("t__A");
		for (IHDSNode task : n.getNodes("/t__A/t"))
		{
			String epId=(String)task.getFirstValue("e");
			String protoId=(String)task.getFirstValue("p");
			String grammarString=(String)task.getFirstValue("grammar");
			if (grammarString!=null)
			{	Document grammarDoc=aContext.transrept(grammarString, Document.class);
				IMessages messages=aContext.getKernelContext().getKernel().getLogger();
				EndpointMetaBuilder epm=new EndpointMetaBuilder();
				MetadataUtils.parseGrammar(grammarDoc.getDocumentElement(), epm, messages);
				IEndpointMeta meta=epm.getMeta(messages);
				EndpointMetaWrapper emw=new EndpointMetaWrapper(meta, null, aContext);
				IHDSNode inputs=task.getFirstNode("i__A");
				IHDSNode v=emw.validateInputs(inputs, messages);
				b.pushNode("t");
				b.addNode("e", epId);
				b.importChildren(v);
				b.popNode();
			}
			else
			{	
				if (protoId!=null)
				{	
					IHDSNode inputs=task.getFirstNode("i__A");
					IHDSNode v=rep.getPaletteItem(protoId,aContext).validateInputs(inputs, aContext);
					b.pushNode("t");
					b.addNode("p", protoId);
					b.addNode("e", epId);
					b.importChildren(v);
					b.popNode();
				}
			}
		}
		
		JSONObject jsonResult=Utils.HDSToJSON(b.getRoot());	
		INKFResponse resp=aContext.createResponseFrom(jsonResult.toString());
		resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
	}
	
	public void onValidateValue(INKFRequestContext aContext, String aEntity) throws Exception
	{	String value=aContext.source("httpRequest:/param/value", String.class);
		String id=aContext.source("httpRequest:/param/id", String.class);
		INKFRequest req=aContext.createRequest("active:nCoDEPalette");
		req.addArgument("entity", aEntity);
		JSONObject json=aContext.source("httpRequest:/param/dynamicImports",JSONObject.class);
		if (json!=null)
		{	req.addArgumentByValue("imports",Utils.JSONToHDS(json).getFirstNode("*"));
		}
		PaletteRepresentation rep=(PaletteRepresentation)aContext.issueRequest(req);
		IHDSNode hds=rep.getPaletteItem(id,aContext).validateValue(value);
		JSONObject jsonResult=Utils.HDSToJSON(hds);	
		aContext.createResponseFrom(jsonResult.toString());
	}
	
	public void onAvailablePalette(INKFRequestContext aContext, String aEntity) throws Exception
	{
		IRepDeployedModules modules=(IRepDeployedModules)aContext.source("active:moduleList",IRepDeployedModules.class);
		NCODEPrototype endpoint=(NCODEPrototype)modules.componentForHash(aEntity);
		if (endpoint==null)
		{	throwNoInstanceException(aContext,aEntity);
		}
		final Set<ISpace> spaces=endpoint.findSpacesInScope(aContext,modules);

		AvailablePalettesRepresentation availablePalettes=aContext.source("active:availablePalettes",AvailablePalettesRepresentation.class);
		IHDSNodeList allPaletteSpaces=availablePalettes.getSpaces();
		
		HDSBuilder b=new HDSBuilder();
		IKernel k=aContext.getKernelContext().getKernel();
		for (IHDSNode spaceNode : allPaletteSpaces)
		{	b.pushNode("palette");
			IIdentifier spaceId=new SimpleIdentifierImpl(spaceNode.getFirstValue("id").toString().trim());
			Version spaceVersion=new Version(spaceNode.getFirstValue("version").toString().trim());
			ISpace space=k.getSpace(spaceId,spaceVersion,spaceVersion);
			String hash=modules.hashForComponent(space);
			IMetaRepresentation meta=modules.metadataForHash(hash);
			b.addNode("hash",hash);
			b.addNode("name",meta.getName());
			b.addNode("id",spaceId.toString());
			b.addNode("version",spaceVersion.toString());
			b.addNode("inScope",spaces.contains(space));
			b.popNode();
		}
		
		//sort
		HDSBuilder b2=new HDSBuilder();
		b2.addNode("supportsDynamicImports", endpoint.getDataFlowConfig(aContext).supportsDynamicImports());
		b2.pushNode("palette__A");
		Comparator<IHDSNode> sorter=new Comparator<IHDSNode>()
		{
			public int compare(IHDSNode o1, IHDSNode o2)
			{	String name1=o1.getFirstValue("name").toString();
				String name2=o2.getFirstValue("name").toString();
				return name1.compareTo(name2);
			}
		};
		for (IHDSNode n: b.getRoot().getNodes("/palette").sort(sorter))
		{	b2.importNode(n);
		}
		
		
		JSONObject jsonResult=Utils.HDSToJSON(b2.getRoot());
		INKFResponse resp=aContext.createResponseFrom(jsonResult.toString());
	}
}